/**
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */
package org.sdm.spa;

import java.util.StringTokenizer;
import java.util.Vector;

import ptolemy.actor.TypedAtomicActor;
import ptolemy.actor.TypedIOPort;
import ptolemy.actor.lib.Const;
import ptolemy.data.BooleanToken;
import ptolemy.data.StringToken;
import ptolemy.data.IntToken;
import ptolemy.data.expr.Parameter;
import ptolemy.data.type.BaseType;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import ptolemy.kernel.util.Workspace;

//////////////////////////////////////////////////////////////////////////
//// EnumItems
/**
 * This Actor extracts items from a comma separated list.
 */

public class EnumItems
    extends TypedAtomicActor {

  /** Constructor
   *  Initialize all three ports.
   *  @param container The container.
   *  @param name The name of this actor.
   *  @exception IllegalActionException If the entity cannot be contained
   *   by the proposed container.
   *  @exception NameDuplicationedException If the container already has an
   *   actor with this name.
   *
   */
  public EnumItems(CompositeEntity container, String name) throws
      NameDuplicationException, IllegalActionException {
    super(container, name);

    portInput = new TypedIOPort(this, "input", true, false);
    portInput.setTypeEquals(BaseType.STRING);

    portOutput = new TypedIOPort(this, "output", false, true);
    portOutput.setTypeEquals(BaseType.STRING);

    portSize = new TypedIOPort(this, "size", false, true);
    portSize.setTypeEquals(BaseType.INT);

    _attachText(
        "_iconDescription",
        "<svg>\n"
        + "<rect x=\"0\" y=\"0\" "
        + "width=\"60\" height=\"30\" "
        + "style=\"fill:white\"/>\n"
        + "</svg>\n");
  }

  ///////////////////////////////////////////////////////////////////
  ////                     ports and parameters                  ////

  public TypedIOPort portInput;
  public TypedIOPort portOutput;
  public TypedIOPort portSize;

  ///////////////////////////////////////////////////////////////////
  ////                         public methods                    ////

  /** Clone the actor into the specified workspace. This calls the
   *  base class and then sets the value public variable in the new
   *  object to equal the cloned parameter in that new object.
   *  @param workspace The workspace for the new object.
   *  @return A new actor.
   *  @exception CloneNotSupportedException If a derived class contains
   *   an attribute that cannot be cloned.
   */
  public Object clone(Workspace workspace) throws CloneNotSupportedException {
    Const newObject = (Const)super.clone(workspace);
    // Set the type constraint.
    newObject.output.setTypeAtLeast(newObject.value);
    return newObject;
  }

  /** Post fire the actor. Return false to indicated that the
   *  process has finished. If it returns true, the process will
   *  continue indefinitely.
   */
  public boolean postfire() throws IllegalActionException {
    return true;
  }

  /** Split the value in the <i>input</i> port, and send each one
   *  to the <i>output</i> port individually.
   *  @exception IllegalActionException If it is thrown by the
   *   send() method sending out the token.
   */
  public void fire() throws IllegalActionException {
    // get input
    String strInput = ( (StringToken) (portInput.get(0))).stringValue();
    _debug("Broadcast: input = " + strInput);

    // get rid of nonword characters in the beginning
    strInput = strInput.replaceAll("^[^\\w]+", "");

    // split and extract all sequences of word characters
    String[] inputs = strInput.split("[^\\w]+");
    portSize.broadcast(new IntToken(inputs.length));
    _debug("Broadcast: size = " + inputs.length);

    // broadcast tokens
    for (int i = 0; i < inputs.length; i++) {
      portOutput.broadcast(new StringToken(inputs[i]));
      _debug("Broadcast: output = " + inputs[i]);
    }
  }
}