/**

 *    '$RCSfile: GrassWebServiceImpl.java,v $'

 *    Copyright: 2004 Regents of the University of California and the

 *              National Center for Ecological Analysis and Synthesis

 *    Authors: @authors@

 *    Release: @release@

 *

 *   '$Author: berkley $'

 *   '$Date: 2004/05/03 22:14:12 $'

 *   '$Revision: 1.17 $'

 *

 * This program is free software; you can redistribute it and/or modify

 * it under the terms of the GNU General Public License as published by

 * the Free Software Foundation; either version 2 of the License, or

 * (at your option) any later version.

 *

 * This program is distributed in the hope that it will be useful,

 * but WITHOUT ANY WARRANTY; without even the implied warranty of

 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the

 * GNU General Public License for more details.

 *

 * You should have received a copy of the GNU General Public License

 * along with this program; if not, write to the Free Software

 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

 *

 */

package org.ecoinformatics.seek.webservice.gis.grass;



import org.ecoinformatics.seek.webservice.gis.grass.GrassWebServiceException;



import java.util.*;

import java.io.*;





/**

 * Implementation of the webservice methods defined in the

 * GrassWebServiceInterface.

 */

public class GrassWebServiceImpl implements GrassWebServiceInterface

{

  /**

   * imports the given arc file into the grass database and returns true

   * if the operation completed successfully.

   * @param arcFileContents the String representation of the contents of the

   * arc file

   * @throws GrassWebServiceException if the service does not complete properly

   * @return boolean true if the service completed successfully

   */

  public boolean importArcToGrass(String arcFileContents, String objectName)

    throws GrassWebServiceException

  {

    String errorout = "";

    try

    {

      ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

      String importCommand = resources.getString("rasterImportCommand");

      System.out.println("running " + importCommand +

        " in method importArcToGrass");

      String tempFile = resources.getString("tempFile");

      String envvars[] = getEnvArray();

      String exe = resources.getString("GISBASE");



      //write out the arcFileContents to a temp file

      writeFile(arcFileContents, new File(tempFile));

      final Process listener;



      //setup the arguments and the environment

      String args[] =

      {

        exe + "/bin/" + importCommand,

         "input=" + tempFile, "output=" + objectName

      };



      //run the command

      listener = Runtime.getRuntime().exec(args, envvars);



      //get the output

      errorout = getErrOutput(listener);



      waitFor(listener);

    }

    catch (Exception e)

    {

      e.printStackTrace();

      throw new GrassWebServiceException("Unexpected error: " + e.getMessage());

    }



    //check for success

    if(errorout.indexOf("CREATING SUPPORT FILES FOR") == -1)

    {

      throw new GrassWebServiceException("Error in importArcToGrass: " +

        errorout);

    }



    return true;

  }



  /**

   * exports and object from the Grass database into an arc file.  This returns

   * the contents of the arc file as a String

   * @param grassObjectName the name of the object in the grass DB to export

   * @return the arc file contents as a String, null if there was an error

   * @throws GrassWebServiceException if the service does not complete properly

   */

  public String exportGrassToArcFile(String grassObjectName)

    throws GrassWebServiceException

  {

    String outFileStr = "";

    String errorout = "";

    String tempFile = "";

    File temp;



    try

    {

      ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

      final Process listener;



      String exportCommand = resources.getString("rasterExportCommand");

      System.out.println("running " + exportCommand +

        " in method exportGrassToArcFile");

      tempFile = resources.getString("tempFile");

      temp = new File(tempFile);

      temp.delete();

      String envvars[] = getEnvArray();

      String exe = resources.getString("GISBASE");



      String args[] = {exe + "/bin/" + exportCommand,

        "input=" + grassObjectName, "output=" + tempFile};



      listener = Runtime.getRuntime().exec(args, envvars);



      //get the output

      errorout = getErrOutput(listener);



      waitFor(listener);

    }

    catch (Exception e)

    {

      e.printStackTrace();

      throw new GrassWebServiceException("Unexpected error: " + e.getMessage());

    }



    //check for success

    if(temp.exists())

    {

      try

      {

        FileReader fr = new FileReader(temp);

        StringBuffer sb = new StringBuffer();

        char[] buff = new char[1024];

        int numread = fr.read(buff, 0 ,1024);

        while(numread != -1)

        {

          sb.append(buff, 0, numread);

          numread = fr.read(buff, 0, 1024);

        }

        return sb.toString();

      }

      catch(Exception e)

      {

        e.printStackTrace();

        throw new GrassWebServiceException("Error writing output file: " +

          e.getMessage());

      }

    }

    else

    {

      throw new GrassWebServiceException("Error in exportGrassToArcFile: " +

        errorout);

    }

  }



  /**

   * Convert an ASCII listing of site locations into a GRASS site list file.

   * @param sites the name of the site object that the inputFile should be

   * imported into

   * @param inputFile the contents of the ascii file that should be imported

   * @param dimensions the number of dimensions represented in the file.

   * @param delimeter either 'character', 'space' or 'tab'

   * @throws GrassWebServiceException if the service does not complete properly

   * @return boolean true if the service completed successfully

   */

  public boolean importAsciiData(String sites, String inputFile,

    int dimensions, String delimiter) throws GrassWebServiceException

  {

    String errorout = "";

    String tempFile = "";

    File temp;



    try

    {

      ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

      final Process listener;



      String command = resources.getString("asciiImportCommand");

      System.out.println("running " + command +

        " in method importAsciiData");

      tempFile = getTempFileName();

      temp = new File(tempFile);

      temp.delete();

      writeFile(inputFile, temp);

      String envvars[] = getEnvArray();

      String exe = resources.getString("GISBASE");



      String args[] = {exe + "/bin/" + command,

        "sites=" + sites, "input=" + tempFile, "d=" + dimensions,

        "fs=" + delimiter};



      listener = Runtime.getRuntime().exec(args, envvars);



      //get the output

      errorout = getErrOutput(listener);



      waitFor(listener);

    }

    catch (Exception e)

    {

      e.printStackTrace();

      throw new GrassWebServiceException("Unexpected error: " + e.getMessage());

    }



    System.out.println("ERROR: " + errorout);

    if(errorout.indexOf("invalid") != -1)

    {

      throw new GrassWebServiceException("Error importing ascii data: " +

        errorout);

    }



    return true;

  }



  /**

   * reports attribute, label, and other information about a sites file

   * @param sites the name of the site object to get info for

   * @return String the info returned on the ascii data

   * @throws GrassWebServiceException if the service does not complete properly

   */

  public String asciiDataInfo(String sites) throws GrassWebServiceException

  {

    String errorout = "";

    String stdout = "";

    try

    {

      ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

      final Process listener;



      String command = resources.getString("asciiInfoCommand");

      System.out.println("running " + command +

        " in method asciiDataInfo");

      String envvars[] = getEnvArray();

      String exe = resources.getString("GISBASE");



      String args[] = {exe + "/bin/" + command,

        "sites=" + sites};



      listener = Runtime.getRuntime().exec(args, envvars);



      //get the output

      errorout = getErrOutput(listener);

      stdout = getStdOutput(listener);



      waitFor(listener);

    }

    catch (Exception e)

    {

      e.printStackTrace();

      throw new GrassWebServiceException("Unexpected error: " + e.getMessage());

    }



    if(errorout.indexOf(sites + " read.") != -1)

    {

      System.out.println("ERROROUT: " + errorout);

      return stdout;

    }

    else

    {

      throw new GrassWebServiceException("Error getting ascii info: " +

        errorout);

    }

  }



  /**

   * Outputs basic information about a user-specified raster map layer.

   * @param map the name of the map object to get info for

   * @return String the info on the raster data

   * @throws GrassWebServiceException if the service does not complete properly

   */

  public String rasterDataInfo(String map) throws GrassWebServiceException

  {

    String errorout = "";

    String stdout = "";

    try

    {

      ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

      final Process listener;



      String command = resources.getString("rasterInfoCommand");

      System.out.println("running " + command +

        " in method rasterDataInfo");

      String envvars[] = getEnvArray();

      String exe = resources.getString("GISBASE");



      String args[] = {exe + "/bin/" + command,

        "map=" + map};



      listener = Runtime.getRuntime().exec(args, envvars);



      //get the output

      errorout = getErrOutput(listener);

      stdout = getStdOutput(listener);



      waitFor(listener);

    }

    catch (Exception e)

    {

      e.printStackTrace();

      throw new GrassWebServiceException("Unexpected error: " + e.getMessage());

    }



    return stdout;

  }



  /**

   * Outputs basic information about a user-specified vector map layer

   * @param vector the name of the vector object to get info for

   * @return String the info on the vector data

   * @throws GrassWebServiceException if the service does not complete properly

   */

  public String vectorDataInfo(String vector) throws GrassWebServiceException

  {

    String errorout = "";

    String stdout = "";

    try

    {

      ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

      final Process listener;



      String command = resources.getString("vectorInfoCommand");

      System.out.println("running " + command +

        " in method vectorDataInfo");

      String envvars[] = getEnvArray();

      String exe = resources.getString("GISBASE");



      String args[] = {exe + "/bin/" + command,

        "input=" + vector};



      listener = Runtime.getRuntime().exec(args, envvars);



      //get the output

      errorout = getErrOutput(listener);

      stdout = getStdOutput(listener);



      waitFor(listener);

    }

    catch (Exception e)

    {

      e.printStackTrace();

      throw new GrassWebServiceException("Unexpected error: " + e.getMessage());

    }



    return stdout;

  }



  /**

   * Uses a GRASS sites list to produce a convex hull vector map

   * @param sites content of sites file to be input

   * @param vect name of vector object to be created

   * @throws GrassWebServiceException if the service does not complete properly

   * @return boolean true if the service completed successfully

   */

  public boolean createHull(String sites, String vect)

    throws GrassWebServiceException

  {

    String errorout = "";

    String stdout = "";

    try

    {

      ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

      final Process listener;



      String command = resources.getString("asciiHullCommand");

      System.out.println("running " + command +

        " in method createHull");

      String envvars[] = getEnvArray();

      String exe = resources.getString("GISBASE");



      String args[] = {exe + "/bin/" + command,

        "sites=" + sites, "vect=" + vect};



      listener = Runtime.getRuntime().exec(args, envvars);



      //get the output

      errorout = getErrOutput(listener);

      stdout = getStdOutput(listener);



      waitFor(listener);

    }

    catch (Exception e)

    {

      e.printStackTrace();

      throw new GrassWebServiceException("Unexpected error: " + e.getMessage());

    }



    if(errorout.indexOf("ERROR") != -1)

    {

      throw new GrassWebServiceException("Error creating hull: " + errorout);

    }



    return true;

  }



  /**

   * Creates GRASS support files for (binary) GRASS vector data.

   * @param vect name of vector object for which to create support objects

   * @throws GrassWebServiceException if the service does not complete properly

   * @return boolean true if the service completed successfully

   */

  public boolean createVectorSupportObjects(String vect)

    throws GrassWebServiceException

  {

    String errorout = "";

    String stdout = "";

    try

    {

      ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

      final Process listener;



      String command = resources.getString("vectorSupportCommand");

      System.out.println("running " + command +

        " in method createVectorSupportObjects");

      String envvars[] = getEnvArray();

      String exe = resources.getString("GISBASE");



      String args[] = {exe + "/bin/" + command,

        "map=" + vect};



      listener = Runtime.getRuntime().exec(args, envvars);



      //get the output

      errorout = getErrOutput(listener);

      stdout = getStdOutput(listener);



      waitFor(listener);

    }

    catch (Exception e)

    {

      e.printStackTrace();

      throw new GrassWebServiceException("Unexpected error: " + e.getMessage());

    }



    if(errorout.indexOf("ERROR") != -1)

    {

      throw new GrassWebServiceException("Error in createVectorSupportObjects: " +

        errorout);

    }



    return true;

  }



  /**

   * Converts a binary GRASS vector map layer into a GRASS raster map layer.

   * @param vect name of vector object to create a raster layer of

   * @param rast name of raster object to create

   * @throws GrassWebServiceException if the service does not complete properly

   * @return boolean true if the service completed successfully

   */

  public boolean convertVectorToRaster(String vect, String rast)

    throws GrassWebServiceException

  {

    String errorout = "";

    String stdout = "";

    try

    {

      ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

      final Process listener;



      String command = resources.getString("vectorToRasterCommand");

      System.out.println("running " + command +

        " in method convertVectorToRaster");

      String envvars[] = getEnvArray();

      String exe = resources.getString("GISBASE");



      String args[] = {exe + "/bin/" + command,

        "input=" + vect, "output=" + rast};



      listener = Runtime.getRuntime().exec(args, envvars);



      //get the output

      errorout = getErrOutput(listener);

      stdout = getStdOutput(listener);



      waitFor(listener);

    }

    catch (Exception e)

    {

      e.printStackTrace();

      throw new GrassWebServiceException("Unexpected error: " + e.getMessage());

    }



    if(errorout.indexOf("ERROR") != -1)

    {

      throw new GrassWebServiceException("Error in convertVectorToRaster: " +

        errorout);

    }



    return true;

  }



  /**

   * Calculates a map based on the product of two rasters.  This method makes

   * the grass call 'r.mapcalc result=rast1*rast2'

   * @param rast1 the first raster input

   * @param rast2 the second raster input

   * @throws GrassWebServiceException if the service does not complete properly

   * @return boolean true if the service completed successfully

   */

  public boolean calculateMap(String rast1, String rast2)

    throws GrassWebServiceException

  {

    String errorout = "";

    String stdout = "";

    try

    {

      ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

      final Process listener;



      String command = resources.getString("rasterMapCalcCommand");

      System.out.println("running " + command +

        " in method calculateMap");

      String envvars[] = getEnvArray();

      String exe = resources.getString("GISBASE");



      String args[] = {exe + "/bin/" + command,

        "result=" + rast1 + "*" + rast2};



      listener = Runtime.getRuntime().exec(args, envvars);



      //get the output

      errorout = getErrOutput(listener);

      stdout = getStdOutput(listener);



      System.out.println("errorout: " + errorout);

      System.out.println("stdout: " + stdout);



      waitFor(listener);

    }

    catch (Exception e)

    {

      e.printStackTrace();

      throw new GrassWebServiceException("Unexpected error: " + e.getMessage());

    }



    if(errorout.indexOf("invalid") != -1)

    {

      throw new GrassWebServiceException("Error in calculateMap: " + errorout);

    }



    return true;

  }



  /**

   * returns the filename of the temp file

   */

  private String getTempFileName()

  {

    //TODO: in the future, when this goes multiuser, we need to add a

    //scheme to make each request to this method unique so that different

    //web service requests don't overwrite the same temp file.

    ResourceBundle resources = ResourceBundle.getBundle("grassconfig");

    return resources.getString("tempFile");

  }



  /**

   * gets the error stream from the process and returns it as a string

   * @param p the process to get the stream from

   * @return String the output of the process

   */

  private static String getErrOutput(Process p) throws IOException

  {

    String errorout = "";

    BufferedReader br_err_in = new BufferedReader(

     new InputStreamReader(p.getErrorStream()));

    String buff = null;

    while ((br_err_in != null && (buff = br_err_in.readLine()) != null))

    {

      errorout += buff + "\n";

    }



    br_err_in.close();

    return errorout;

  }



  /**

   * gets the std out stream from the process and returns it as a string

   * @param p the process to get the stream from

   * @return String the output of the process

   */

  private static String getStdOutput(Process p) throws IOException

  {

    String out = "";

    String buff;

    BufferedReader br_in = new BufferedReader(

     new InputStreamReader(p.getInputStream()));

    while ((br_in != null && (buff = br_in.readLine()) != null))

    {

      out += buff + "\n";

    }



    br_in.close();

    return out;

  }



   /**

   * writes the arcfile contents string to a temp file for use with the

   * exec call

   */

  private static void writeFile(String fileContents, File tempFile)

    throws Exception

  {

    FileWriter fw = new FileWriter(tempFile);

    fw.write(fileContents, 0, fileContents.length());

    fw.flush();

    fw.close();

  }



  /**

   * creates the array of needed env variables for grass

   */

  private static String[] getEnvArray()

  {

    //get grass params from the property bundle

    ResourceBundle resources = ResourceBundle.getBundle("grassconfig");



    String HOME = resources.getString("HOME");

    String GISBASE = resources.getString("GISBASE");

    String LOCATION_NAME = resources.getString("LOCATION_NAME");

    String GISDBASE = resources.getString("GISDBASE");

    String GISRC = resources.getString("GISRC");



    String envvars[] =

    {

      "home" + HOME,

      "HOME" + HOME,

      "Home" + HOME,

      "LOCATION_NAME=" + LOCATION_NAME,

      "GISBASE=" + GISBASE,

      "GISDBASE=" + GISDBASE,

      "GISRC=" + GISRC,

    };



    return envvars;

  }





  /**

   * private method that waits for a process to finish before continuing

   * @param listener the process to wait for

   */

  private void waitFor(final Process listener)

  {

    //thread to wait for grass process to terminate

    new Thread(new Runnable()

    {

     public void run()

     {

       try

       {

         listener.waitFor();

       }

       catch (InterruptedException ie)

       {

         ie.printStackTrace();

         System.out.println("InterruptedException caught while waiting " +

           "for Grass Process to exit: " + ie);

       }

     }

    }).start();

  }

}

