/**
 *    '$RCSfile: DBTablesGenerator.java,v $'
 *
 *     '$Author: tao $'
 *       '$Date: 2004/08/12 21:13:09 $'
 *   '$Revision: 1.5 $'
 *
 *  For Details: http://kepler.ecoinformatics.org
 *
 * Copyright (c) 2004 The Regents of the University of California.
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY
 * FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
 * ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN
 * IF THE UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY
 * OF CALIFORNIA HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT,
 * UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 */

package org.ecoinformatics.seek.dataquery;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Timestamp;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.ecoinformatics.seek.datasource.Attribute;
import org.ecoinformatics.seek.datasource.TableEntity;
import org.ecoinformatics.seek.datasource.UnresolvableTypeException;
import org.ecoinformatics.util.DebugPrinter;
import org.ecoinformatics.util.Util;



/**
 * This class will generate db tables base on a given TableEntity which was
 *  created by a metadata parser.
 * @author Jing Tao
 *
 */

public class DBTablesGenerator
{
  private TableEntity table = null;
  private String tableName =  null;
  private Vector dbJavaDataTypeList = new Vector();
  private DebugPrinter dbg = DebugPrinter.getInstance();

  // Constant
  private static final String CREATETEXTTABLE = "CREATE TEXT TABLE";
  private static final String CREATETABLE     = "CREATE TABLE";
  private static final String LEFTPARENTH     = "(";
  private static final String RIGHTPARENTH    = ")";
  private static final String SEMICOLON       = ";";
  private static final String SPACE           = " ";
  private static final String COMMA           = ",";
  private static final String QUOTE           = "\"";
  private static final String FIELDSEPATATOR  = "fs=";
  private static final String INSERT          = "INSERT INTO";
  private static final String VALUES          = "VALUES";
  private static final String QUESTION        = "?";
  private static final String SETTABLE        = "SET TABLE";
  private static final String SOURCE          = "SOURCE";
  private static final String IGNOREFIRST     ="ignore_first=true";
  private static final String STRING          = "String";
  private static final String INTEGER         = "Integer";
  private static final String LONG            = "Long";
  private static final String DOUBLE          = "Double";
  private static final String FLOAT           = "Float";
  private static final String DATETIME        = "Timestamp";
  private static final String BOOLEAN         = "Boolean";

  /**
   * Constructor of DBTablesGenerator. This is for non-text file tables. And will
   * load data into db.
   * @param givenEntityList Hashtable
   * @param givenData Vector[]  This a vector array, each vector is one row of
   * data. The element in vector is a string
   */
  public DBTablesGenerator(TableEntity givenEntityList, Vector[] givenData )
                          throws SQLException, UnresolvableTypeException, ClassNotFoundException
  {
    table = givenEntityList;
    if (table == null)
    {
      dbg.print("The entity is null and couldn't create table for it", 2);
      throw new SQLException("The entity is null and couldn't create table for it");
    }
    tableName = table.getName();
    if (tableName == null || tableName.trim().equals(""))
    {
      dbg.print("There is no table name in entity", 2);
      throw new SQLException("Couldn't create a table because not tablename " +
                            "is given");
    }
    // this is for non-text type table
    String generateTalbeSql = generateDDLForOneEntity(CREATETABLE);
    excuteSQLCommand(generateTalbeSql);
    loadDataIntoTable(givenData);

  }

  /**
  * Constructor of DBTablesGenerator. This is for text file tables. And will
  * load data into db.
  * @param givenEntityList Hashtable
  * @param givenData Vector[]  This a vector array, each vector is one row of
  * data. The element in vector is a string
  */
 public DBTablesGenerator(TableEntity givenEntityList, String textFileLocation)
                         throws SQLException, UnresolvableTypeException, ClassNotFoundException
 {
   table = givenEntityList;
   if (table == null)
   {
     dbg.print("The entity is null and couldn't create table for it", 2);
     throw new SQLException("The entity is null and couldn't create table for it");
   }
   tableName = table.getName();
   if (tableName == null || tableName.trim().equals(""))
   {
     dbg.print("There is no table name in entity", 2);
     throw new SQLException("Couldn't create a table because not tablename " +
                           "is given");
   }
   int numOfHeadLines = table.getNumHeaderLines();
   // hsql only handle two scenaro no head line or one head line
   boolean ignoreHeadLines = false;
   if (numOfHeadLines == 0)
   {
      ignoreHeadLines = false;
   }
   else if (numOfHeadLines == 1)
   {
     ignoreHeadLines = true;
   }
   else
   {
       throw new SQLException("HSQL text table only handle one line header" +
                              " and this entity has " + numOfHeadLines +
                              " headlines");
   }
   // if this is attribute row oriented, hsql can't handle it
   String orientation = table.getOrientation();
   if (orientation != null && orientation.equals(TableEntity.ROWMAJOR))
   {
     throw new SQLException("DB doesn't handle a text table which attribute is"+
                            " row oriented");
   }
   // this is for text type table
   String generateTableSql = generateDDLForOneEntity(CREATETEXTTABLE);
   excuteSQLCommand(generateTableSql);

   // bind text source to table
   String delimiterStr = table.getDelimiter();
   // need to figure out the delimiter str in db. The format is different
   DelimiterResolver resolver = new DelimiterResolver();
   String dbDelimiter = resolver.resolve(delimiterStr);
   String bindTextFileToTalbeSql =
                         generateBindTextFileToTableSQL(textFileLocation,
                                            dbDelimiter, ignoreHeadLines);
   excuteSQLCommand(bindTextFileToTalbeSql);
 }


  /*
   * Create a table base one given DDL
   */
  private void excuteSQLCommand(String sql) throws SQLException, ClassNotFoundException
  {
    Connection conn = Util.getDBConnection();
    Statement st = conn.createStatement();
    st.execute(sql);
    st.close();
    conn.close();
  }//generateTable

  /*
   * Method to load data into table. If error happend, it will roll back.
   * Vector is String vector in vector array data.
   */
  private void loadDataIntoTable(Vector[] data) throws SQLException,
    ClassNotFoundException, IllegalArgumentException, UnresolvableTypeException
  {
    if (data == null)
    {
      return;
    }

    PreparedStatement pStatement = null;
    Connection conn = Util.getDBConnection();
    conn.setAutoCommit(false);
    try
    {
      String insertCommand = generateInsertCommand();
      pStatement = conn.prepareStatement(insertCommand);
      int length = data.length;
      for (int i=0; i<length; i++)
      {
        Vector row = data[i];
        // insert one row data into table
        int sizeOfRow = row.size();
        for (int j = 0; j < sizeOfRow; j++)
        {
          String dataElement = (String) row.elementAt(j);
          // get data type for the vector which already has the cloumn java
          // type info after parsing attribute in private method parseAttributeList
          String javaType = (String) dbJavaDataTypeList.elementAt(j);
          // this method will binding data into preparedstatement base on
          // java data type
          // The index of pstatement start 1 (Not 0), so it should j+1.
          pStatement = setupPreparedStatmentParameter(j+1, pStatement,
              dataElement, javaType);

        }
        pStatement.execute();
      }
    }
    catch(SQLException sqle)
    {
      conn.rollback();
      pStatement.close();
      conn.close();
      throw sqle;
    }
    catch(IllegalArgumentException le)
    {
      conn.rollback();
      pStatement.close();
      conn.close();
      throw le;
    }
    catch(UnresolvableTypeException ue)
    {
      conn.rollback();
      pStatement.close();
      conn.close();
      throw ue;
    }
    conn.commit();
    pStatement.close();
    conn.close();
  }//loadDataIntoTable

  /*
   * Method for generate sql command to create table
   */
  private String generateDDLForOneEntity(String tableType) throws SQLException,
                                               UnresolvableTypeException
  {
    StringBuffer sql            = new StringBuffer();
    String textFileName   = table.getFileName();
    int    headLineNumber = table.getNumHeaderLines();
    String orientation    = table.getOrientation();
    String delimiter      = table.getDelimiter();
    sql.append(tableType);
    sql.append(SPACE);
    sql.append(tableName);
    sql.append(LEFTPARENTH);
    Attribute[] attributeList = table.getAttributes();
    String attributeSql = parseAttributeList(attributeList);
    sql.append(attributeSql);
    sql.append(RIGHTPARENTH);
    sql.append(SEMICOLON);
    String sqlStr = sql.toString();
    dbg.print("The command to create tables is "+sqlStr, 2);
    return sqlStr;
  }//generateDDLForOneEntity
  /*
   * Add attribute defination in create table command. If one attribute is null
   * or has same error an exception will be throw
   */
  private String parseAttributeList(Attribute[] list) throws SQLException,
                                                   UnresolvableTypeException
  {
    StringBuffer attributeSql = new StringBuffer();
    if (list == null || list.length == 0)
    {
      dbg.print("There is no attribute defination in entity", 2);
      throw new SQLException("There is no attribute defination in entity");
    }
    int size = list.length;
    DBDataTypeResolver dataTypeResolver = new DBDataTypeResolver();
    boolean firstAttribute = true;
    for (int i = 0; i < size; i++)
    {
      Attribute attribute = list[i];
      if (attribute == null)
      {
        dbg.print("One attribute defination is null attribute list", 2);
        throw new SQLException("One attribute defination is null attribute list");
      }
      String name = attribute.getName();
      String dataType = attribute.getDataType();
      String dbDataType = dataTypeResolver.resolveDBType(dataType);
      String javaDataType = dataTypeResolver.resolveJavaType(dataType);
      dbJavaDataTypeList.add(javaDataType);
      if (!firstAttribute)
      {
        attributeSql.append(COMMA);
      }
      attributeSql.append(name);
      attributeSql.append(SPACE);
      attributeSql.append(dbDataType);
      firstAttribute = false;

    }//for
    return attributeSql.toString();
  }//parseAttributeList

  /*
   * Generate a sql command to for insert data into talbe. Here we use
   * PreparedStatement.
   */
  private String generateInsertCommand() throws SQLException
  {
    StringBuffer sql = new StringBuffer();
    sql.append(INSERT);
    sql.append(SPACE);
    sql.append(tableName);
    sql.append(LEFTPARENTH);
    Attribute[] list = table.getAttributes();
    if (list == null || list.length == 0)
    {
      dbg.print("There is no attribute defination in entity", 2);
      throw new SQLException("There is no attribute defination in entity");
    }
    int size = list.length;
    // cloumna name part
    boolean firstAttribute = true;
    for (int i = 0; i< size; i++)
    {
      Attribute attribute = list[i];
      if (attribute == null)
      {
       dbg.print("One attribute defination is null attribute list", 2);
       throw new SQLException("One attribute defination is null attribute list");
      }
      String name = attribute.getName();
      if (!firstAttribute)
      {
        sql.append(COMMA);
      }
      sql.append(name);
      firstAttribute = false;
    }
    sql.append(RIGHTPARENTH);
    sql.append(SPACE);
    sql.append(VALUES);
    sql.append(SPACE);
    sql.append(LEFTPARENTH);
    //value part, use ? replace
    firstAttribute = true;
    for (int i = 0; i <size; i++)
    {
      if (!firstAttribute)
      {
        sql.append(COMMA);
      }
      sql.append(QUESTION);
      firstAttribute = false;
    }
    sql.append(RIGHTPARENTH);
    sql.append(SEMICOLON);
    dbg.print("The insert command is " +sql.toString(), 2);
    return sql.toString();
  }

  /*
   * Method to setup data for prepare statment
   */
  private PreparedStatement setupPreparedStatmentParameter(int index,
                 PreparedStatement pStatement,
                 String data, String javaDataType) throws SQLException,
                  UnresolvableTypeException, IllegalArgumentException
  {
    if (pStatement == null )
    {
      return pStatement;
    }

    // set default type as string
    if (javaDataType == null )
    {
      pStatement.setString(index, data);
    }
    else
    {
        if ( javaDataType.equals(STRING))
        {
          pStatement.setString(index, data);
        }
        else if ( javaDataType.equals(INTEGER))
        {
          pStatement.setInt(index, (new Integer(data)).intValue());
        }
        else if (javaDataType.equals(DOUBLE))
        {
          pStatement.setDouble(index, (new Double(data)).doubleValue());
        }
        else if (javaDataType.equals(FLOAT))
        {
          pStatement.setFloat(index, (new Float(data)).floatValue());
        }
        else if (javaDataType.equals(BOOLEAN))
        {
          pStatement.setBoolean(index, (new Boolean(data)).booleanValue());
        }
        else if (javaDataType.equals(LONG))
        {
          pStatement.setLong(index, (new Long(data)).longValue());
        }
        else if (javaDataType.equals(DATETIME))
        {
          pStatement.setTimestamp(index, Timestamp.valueOf(data));
        }
        else
        {
            throw new  UnresolvableTypeException("This java type " + javaDataType +
                     " has NOT implement in " +
                     "DBTablesGenerator.setupPreparedStatmentParameter method");
        }
    }
    return pStatement;
  }//setupPreparedStatmentParameter

  private String generateBindTextFileToTableSQL(String textFilePath,
                                      String delimiter, boolean ignoreFirstLine)
                                                throws SQLException
  {
    if (textFilePath == null || textFilePath.trim().equals(""))
    {
      dbg.print("No file location specify for this text table", 2);
      throw new SQLException("No file location specify for this text table");
    }
    if (delimiter == null)
    {
      throw new SQLException("No delimiter be specified in metadata");
    }
    StringBuffer sql = new StringBuffer();
    sql.append(SETTABLE);
    sql.append(SPACE);
    sql.append(tableName);
    sql.append(SPACE);
    sql.append(SOURCE);
    sql.append(SPACE);
    sql.append(QUOTE);
    sql.append(textFilePath);
    sql.append(SEMICOLON);
    // delimiter part
    sql.append(FIELDSEPATATOR);
    sql.append(delimiter);

    if (ignoreFirstLine)
    {
      sql.append(SEMICOLON);
      sql.append(IGNOREFIRST);
    }
    sql.append(QUOTE);
    dbg.print("The set source command is " + sql.toString(), 2);
    return sql.toString();
  }
}//DBTablesGenerator
